/*
Copyright 2021 The Knative Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package broker

import (
	"knative.dev/reconciler-test/pkg/feature"

	"knative.dev/eventing/test/rekt/resources/broker"
)

func DefaultDeliverySpec() *feature.Feature {
	f := feature.NewFeatureNamed("default delivery spec")

	brokerName := feature.MakeRandomK8sName("broker")

	f.Setup("install broker", broker.Install(brokerName, broker.WithEnvConfig()...))

	f.Requirement("broker is ready", broker.IsReady(brokerName))
	f.Requirement("broker is addressable", broker.IsAddressable(brokerName))

	f.Assert("broker has delivery spec set", broker.WaitForCondition(brokerName,
		broker.HasDelivery().
			And(broker.HasDeliveryRetry()).
			And(broker.HasDeliveryBackoffDelay()).
			And(broker.HasDeliveryBackoffPolicy()),
	))

	return f
}
